import QuantLib as ql
from typing import List
from .pymodels import *
from .ql_utils import *
from .ql_conventions import *


def deporate_2_depohelpers(depo_setting: DepoSetting, rates: List[Rate],  calendar=None ):
    if not calendar:
        cal = ql.WeekendsOnly()
    else:
        cal = calendar
    depo_helpers = []
    business_day = ql_business_day[depo_setting.business_day]
    day_count = ql_day_count[depo_setting.day_count]
    for rate in rates:
        quote_handle = ql.QuoteHandle(ql.SimpleQuote(rate.rate/100))
        ql_period = ql.Period(rate.tenor)
        dr_helper = ql.DepositRateHelper(quote_handle, ql_period, 
                    depo_setting.start_basis, cal, business_day,
                    depo_setting.eom, day_count)
        depo_helpers.append(dr_helper)
    
    return depo_helpers


def parrate_2_bondhelpers(value_date: str,  
                        parbond_setting: BondSetting, 
                        rates: List[Rate],
                        calendar=None ):
    if not calendar:
        cal = ql.WeekendsOnly()
    else:
        cal = calendar

    vdate = datestr_to_qldate(value_date)    
    ql_period = ql.Period(ql_freq_tenor[parbond_setting.frequency])
    par_value = 100
    quote_handle = ql.QuoteHandle(ql.SimpleQuote(par_value))
    
    fix_helpers = [] #list to keep individual helper
    for rate in rates:
        #Need ql.Schedule to use ql.FixedRateBondHelper
        maturity = cal.advance(vdate, ql.Period(rate.tenor),     
                                    ql_business_day[parbond_setting.business_day], 
                                    True)
        schedule = ql.Schedule(vdate, 
                            maturity, 
                            ql_period, 
                            cal, 
                            ql_business_day[parbond_setting.business_day],
                            ql_business_day[parbond_setting.terminate_business_day],
                            ql_date_generation[parbond_setting.date_gen],
                            parbond_setting.is_eom)
    
        fix_helper = ql.FixedRateBondHelper(quote_handle, 
                            parbond_setting.start_basis, 
                            par_value, 
                            schedule, 
                            [rate.rate/100],
                            ql_day_count[parbond_setting.day_count])
        fix_helpers.append(fix_helper)

    return fix_helpers





from pymodels import *

class FixedBondVal:
    def __init__(self):
        pass

    @property
    def bond_info(self):
        return self.__bond_info

    @bond_info.setter
    def bond_info(self, bond_info):
        self.__bond_info = bond_info
        


from .bonds import get_qlfixedratebond
import QuantLib as ql 
from .ql_enums import *
from .pymodels import *
from .ql_utils import *
from .ql_conventions import *
from .termstructure import *


def flat_forward():
    thedict = {"date": '2021-02-28',
                "rate": 5.00,
                "day_count": DayCount.act_act_act365,
                "frequency": Frequency.semi_annual,
                "compounding": Compounding.compounded }
    ffwd = FlatForwardCurve(**thedict)
    return ffwd


def sample_piecewise(value_date):
    rates, depo_setting, parrates, par_setting = sample_curveinfo()

    return curve_piecewise(value_date, 
                            depo_setting, 
                            rates, 
                            par_setting, 
                            parrates)


def sample_curveinfo():
    depo_rates = [
        {"tenor": "1M", "rate": 2.00},
        {"tenor": "2M", "rate": 2.05}, 
        {"tenor": "3M", "rate": 2.10},
        {"tenor": "6M", "rate": 2.20}
    ]

    rates = []
    for depo_rate in depo_rates:
        rate = Rate(**depo_rate)
        rates.append(rate)

    depo_sett = {
        "start_basis": StartBasis.today, 
        "business_day": BusinessDay.mod_following,
        "day_count": DayCount.act_act_act365,
        "eom": False
        }
    depo_setting = DepoSetting(**depo_sett)

    par_rates = [
        {"tenor": "1Y", "rate": 2.30},
        {"tenor": "2Y", "rate": 2.40}, 
        {"tenor": "3Y", "rate": 2.50},
        {"tenor": "5Y", "rate": 2.70},
        {"tenor": "7Y", "rate": 2.80},
        {"tenor": "10Y", "rate": 2.90},
        {"tenor": "20Y", "rate": 3.00},
        {"tenor": "30Y", "rate": 3.10}
    ]

    parrates = []
    for par_rate in par_rates:
        rate = Rate(**par_rate)
        parrates.append(rate)

    par_sett = {
        "frequency": "Semi-Annual",
        "start_basis": 2,
        "day_count": "Actual/Actual (ISMA)",
        "business_day": "Modified Following",
        "terminate_business_day": "Modified Following",
        "date_gen": "Backward from maturity date",
        "is_eom":  False
    }
    par_setting = BondSetting(**par_sett)

    return rates, depo_setting, parrates, par_setting


def sample_spreadinfo():
    spread_rates = [
        {"tenor": "1M", "rate": 0.4},
        {"tenor": "2M", "rate": 0.45}, 
        {"tenor": "3M", "rate": 0.5},
        {"tenor": "6M", "rate": 0.55},
        {"tenor": "1Y", "rate": 0.6},
        {"tenor": "2Y", "rate": 0.65}, 
        {"tenor": "3Y", "rate": 0.75},
        {"tenor": "5Y", "rate": 0.85},
        {"tenor": "7Y", "rate": 0.95},
        {"tenor": "10Y", "rate": 1.1},
        {"tenor": "20Y", "rate": 1.3},
        {"tenor": "30Y", "rate": 1.5}
    ]
    rates = []
    for depo_rate in spread_rates:
        rate = Rate(**depo_rate)
        rates.append(rate)
    return rates


#NOTE: changing business_day to no_adjustment will not give
#      clean price of 100 even on issue date.
def sample_bondinfo():
    bondsetting = {
        "frequency":Frequency.semi_annual,
        "start_basis": StartBasis.today,
        "day_count": DayCount.act_act_isma,
        "business_day": BusinessDay.no_adjustment,
        "terminate_business_day": BusinessDay.no_adjustment,
        "date_gen": DateGeneration.backward,
        "is_eom":  True
        }

    setting = BondSetting(**bondsetting)

    bonddata = {"issue_date": "2021-08-18", 
                "maturity": "2025-08-18", 
                "settings": setting, 
                "coupon": 5.00,
                "face_value": 10_000_000
                }
    thebond = FixedRateBond(**bonddata)

    return thebond, setting


# NOTE: it is possible but yet to be tested to have multiple coupons
# for a fixed rate bond since the argument accepts list of coupons
def sample_fixedratebond():
    thebond, setting = sample_bondinfo()
    ql_period = ql.Period(ql_freq_tenor[setting.frequency])
    #print(ql_period)
    #1. Create Schedule
    schedule = ql.Schedule(datestr_to_qldate(thebond.issue_date), 
                        datestr_to_qldate(thebond.maturity), 
                        ql_period, 
                        ql.WeekendsOnly(), 
                        ql_business_day[setting.business_day],
                        ql_business_day[setting.terminate_business_day],
                        ql_date_generation[setting.date_gen],
                        setting.is_eom)
    fixbond = ql.FixedRateBond(setting.start_basis, 
                thebond.face_value, 
                schedule, 
                [thebond.coupon/100], 
                ql_day_count[setting.day_count])
    return fixbond


def sample_index(setting: BondSetting):
    # Create and index for Floating Rate
    ff = ql.FlatForward(setting.start_basis, 
                        ql.WeekendsOnly(), 
                        0.05,
                        ql_day_count[setting.day_count],
                        ql.Simple, 
                        ql_frequency[setting.frequency])
    
    #yts = ql.RelinkableYieldTermStructureHandle() 
    #yts.linkTo(ff)
    ff_handle = ql.YieldTermStructureHandle(ff)
    frb_engine = ql.DiscountingBondEngine(ff_handle)

    myr_br_index = ql.IborIndex('MYR-BR', 
                        ql.Period('6M'), 
                        0, 
                        ql.MYRCurrency(), 
                        ql.WeekendsOnly(), 
                        ql.ModifiedFollowing, 
                        True, 
                        ql.ActualActual(ql.ActualActual.Actual365),
                        ff_handle)
    #historical fixing is required for the active coupon period
    myr_br_index.addFixing(ql.Date(5,1,2021),0.05)

    return myr_br_index, frb_engine


def sample_FRB():
    thebond, setting = sample_bondinfo()
    ql_period = ql.Period(ql_freq_tenor[setting.frequency])
    schedule = ql.Schedule(datestr_to_qldate(thebond.issue_date), 
                        datestr_to_qldate(thebond.maturity), 
                        ql_period, 
                        ql.WeekendsOnly(), 
                        ql_business_day[setting.business_day],
                        ql_business_day[setting.terminate_business_day],
                        ql_date_generation[setting.date_gen],
                        setting.is_eom)

    index, engine = sample_index(setting)
    frb = ql.FloatingRateBond(setting.start_basis, 
                            thebond.face_value, 
                            schedule, 
                            index, 
                            ql.ActualActual(ql.ActualActual.Actual365), 
                            spreads=[0.005])
    # Note changing the engine doesnt seem to have an effect
    # presumably bcoz index is set in ql.FloatingRateBond
    frb.setPricingEngine(engine)

    return frb


# NOTE: it is possible but yet to be tested to have multiple coupons
# and face value for a  bond since the argument accepts list of coupons
# and face values. However, the face value can only be amortising which is 
# equivalent to using ql.AmortizingFixedRateBond
def sample_qlbond():
    thebond, setting = sample_bondinfo()
    abond = get_qlfixedratebond(thebond, setting,ql.WeekendsOnly())
    ql_period = ql.Period(ql_freq_tenor[setting.frequency])
    #print(ql_period)
    #1. Create Schedule
    schedule = ql.Schedule(datestr_to_qldate(thebond.issue_date), 
                        datestr_to_qldate(thebond.maturity), 
                        ql_period, 
                        ql.WeekendsOnly(), 
                        ql_business_day[setting.business_day],
                        ql_business_day[setting.terminate_business_day],
                        ql_date_generation[setting.date_gen],
                        setting.is_eom)
    interest = ql.FixedRateLeg(schedule, 
                                ql_day_count[setting.day_count], 
                                [thebond.face_value], 
                                [thebond.coupon/100])

    bond = ql.Bond(0, ql.WeekendsOnly(), 
                    datestr_to_qldate(thebond.issue_date),
                    interest)
    return bond


def sample_bondrisk(thebond, ql_irate):

    clean_price = ql.BondFunctions.cleanPrice(thebond, ql_irate)
    accrued = ql.BondFunctions.accruedAmount(thebond)
    bps = ql.BondFunctions.bps(thebond, ql_irate)
    duration = ql.BondFunctions.duration(thebond, ql_irate, ql.Duration.Macaulay)
    modified_duration = ql.BondFunctions.duration(thebond, ql_irate, ql.Duration.Modified)
    convexity = ql.BondFunctions.convexity(thebond, ql_irate)
    basis_point_value = ql.BondFunctions.basisPointValue(thebond, ql_irate)
    yield_value_basis_point = ql.BondFunctions.yieldValueBasisPoint(thebond, ql_irate)
    stats = {"accrued":  accrued,
            "clean_price": clean_price,
            "bps": bps,
            "duration": duration,
            "modified_duration": modified_duration,
            "convexity": convexity,
            "basis_point_value": basis_point_value,
            "yield_value_basis_point": yield_value_basis_point}
    bondrisk = BondRisk(**stats)
    return bondrisk


def sample_structuredloaninfo():
    loan_setting = {
        "frequency":Frequency.monthly,
        "start_basis": StartBasis.today,
        "day_count": DayCount.act_act_act365,
        "business_day": BusinessDay.mod_following,
        "terminate_business_day": BusinessDay.mod_following,
        "date_gen": DateGeneration.forward,
        "is_eom":  True
    }

    setting = BondSetting(**loan_setting)
    
    loan_data = {"issue_date": "2021-01-05", 
                "maturity": "2024-01-05", 
                "settings": setting, 
                }
    theloan = StructuredBond(**loan_data)

    return theloan, setting


def sample_fixrate_structuredloan():
    theloan, setting  = sample_structuredloaninfo()
    ql_period = ql.Period(ql_freq_tenor[setting.frequency])
    schedule = ql.Schedule(datestr_to_qldate(theloan.issue_date), 
                        datestr_to_qldate(theloan.maturity), 
                        ql_period, 
                        ql.WeekendsOnly(), 
                        ql_business_day[setting.business_day],
                        ql_business_day[setting.terminate_business_day],
                        ql_date_generation[setting.date_gen],
                        setting.is_eom)
    
    # Create the loan rate structure which increase by 0.25%
    # info is used by ql.FixedRateLeg
    loan_rates = []
    rate_initial = 5.00
    step = 0.00
    step_size = 0.25
    for i in range(1, len(schedule)):
        loan_rates.append((rate_initial + step)/100)
        step += step_size

    # Setting the array of face values which increases by 200,000
    # info is used by ql.FixedRateLeg
    face_values = []
    fv_initial = 10_000_000.00
    step = 0.00
    fv_step_size = 200_000.00
    for i in range(1, len(schedule)):
        face_values.append(fv_initial + step)
        step += fv_step_size
    #print(face_values)
    interests = ql.FixedRateLeg(schedule, ql_day_count[setting.day_count], 
                face_values, loan_rates)

    cashflows = []
    no_of_cfs = len(face_values)
    fvflows = [-fv_step_size for i in range(no_of_cfs - 1)]
    #print(no_of_cfs)
    for i in range(no_of_cfs):
        if i == no_of_cfs - 1:
            cashflow = face_values[i] - sum(fvflows) +interests[i].amount()
            #print(interests[i].date(), face_values[i], sum(fvflows), interests[i].amount() )
        else:
            cashflow = fvflows[i] + interests[i].amount()
            #print(interests[i].date(), fvflows[i], interests[i].amount() )
        simple_cash_flow = ql.SimpleCashFlow(cashflow, interests[i].date())
        cashflows.append(simple_cash_flow)

    loanleg = ql.Leg(cashflows)

    return loanleg


def sample_floatingrate_structuredloan():
    theloan, setting  = sample_structuredloaninfo()
    ql_period = ql.Period(ql_freq_tenor[setting.frequency])
    schedule = ql.Schedule(datestr_to_qldate(theloan.issue_date), 
                        datestr_to_qldate(theloan.maturity), 
                        ql_period, 
                        ql.WeekendsOnly(), 
                        ql_business_day[setting.business_day],
                        ql_business_day[setting.terminate_business_day],
                        ql_date_generation[setting.date_gen],
                        setting.is_eom)
    
    
    # Setting the array of face values which increases by 200,000
    # info is used by ql.FixedRateLeg
    face_values = []
    fv_initial = 10_000_000.00
    step = 0.00
    fv_step_size = 200_000.00
    for i in range(1, len(schedule)):
        face_values.append(fv_initial + step)
        step += fv_step_size
    
    index, engine = sample_index(setting)
    loanleg = ql.IborLeg(face_values, 
                schedule, 
                index, 
                ql.Actual360(), 
                ql.ModifiedFollowing, 
                fixingDays=[0])

    return loanleg


def callabilityschedule(qlSchedule: ql.QuantLib.Schedule):
    
    callability_schedule = ql.CallabilitySchedule()
    call_price = 100.0
    no_of_dates = len(list(qlSchedule))
    start = int(no_of_dates/4)
    #print(start, no_of_dates)
    for i in range(start, no_of_dates):
        callability_price  = ql.CallabilityPrice(
            call_price, ql.CallabilityPrice.Clean)
        callability_schedule.append(
                ql.Callability(callability_price, 
                            ql.Callability.Call,
                            qlSchedule[i]))
        print(qlSchedule[i])
    return callability_schedule


def callable_bondinfo():
    bondsetting = {
        "frequency":Frequency.semi_annual,
        "start_basis": StartBasis.today,
        "day_count": DayCount.act_act_isma,
        "business_day": BusinessDay.no_adjustment,
        "terminate_business_day": BusinessDay.following,
        "date_gen": DateGeneration.backward,
        "is_eom":  True
        }

    setting = BondSetting(**bondsetting)

    bonddata = {"issue_date": "2021-08-18", 
                "maturity": "2031-08-18", 
                "settings": setting, 
                "coupon": 5.00,
                "face_value": 10_000_000
                }
    thebond = FixedRateBond(**bonddata)

    return thebond, setting


def curve_with_spread(value_date: str):
    vdate = datestr_to_qldate(value_date) 
    calendar = ql.WeekendsOnly()
    rates, depo_setting, parrates, par_setting = sample_curveinfo()
    t_structure = curve_piecewise(value_date, 
                                depo_setting, 
                                rates,
                                par_setting, 
                                parrates)
    curve_handle = ql.YieldTermStructureHandle(t_structure)
    spreads = sample_spreadinfo()
    spread_handles = []
    spread_dates = []
    spread_quotes = []

    for spread in spreads:
        spread_quote = ql.SimpleQuote(spread.rate/100)
        spread_quotes.append(spread_quote)
        spread_handle = ql.QuoteHandle(spread_quote)
        spread_handles.append(spread_handle)
        maturity = calendar.advance(vdate, get_qlPeriod(spread.tenor))
        spread_dates.append(maturity)

    spread_curve = ql.SpreadedLinearZeroInterpolatedTermStructure(
                            curve_handle,
                            spread_handles,
                            spread_dates)
    spread_curve_handle = ql.YieldTermStructureHandle(spread_curve)
    return spread_curve_handle


def srmodel_info():
    callbond_model = {"name": "HullWhite",
                            "vas_r0": 0.05,
                            "vas_a": 0.1,
                            "vas_b": 0.05,
                            "vas_sigma": 0.01,
                            "vas_lambda": 0.00,
                            "bk_a": 0.1,
                            "bk_sigma": 0.1,
                            "hw_a": 0.1,
                            "hw_sigma": 0.01,
                            "gsr_step": None,
                            "gsr_sigma": None,
                            "gsr_reversion": None,
                            "g2_a": 0.1,
                            "g2_b": 0.1,
                            "g2_sigma": 0.01,
                            "g2_eta": 0.01,
                            "g2_rho": -0.75,
                        }
    return callbond_model


from typing import List
from .pymodels import *
import QuantLib as ql
from .pym_ql_utils import *
from .ql_utils import *
from .ql_conventions import *


def curve_piecewise(value_date: str, 
                depo_setting: DepoSetting, 
                depo_rates: List[Rate], 
                par_setting:BondSetting, 
                par_rates: List[Rate], 
                cal = None, 
                method: PiecewiseMethods = PiecewiseMethods.logcubicdiscount):
    
    vdate = datestr_to_qldate(value_date) 
    ql.Settings.instance().evaluationDate = vdate
    if not cal:
        calendar = ql.WeekendsOnly()
    else:
        calendar = ql.WeekendsOnly()
    
    depo_helpers = deporate_2_depohelpers(depo_setting, depo_rates)
    bond_helpers = parrate_2_bondhelpers(value_date, par_setting, par_rates)

    rate_helpers = depo_helpers + bond_helpers
    
    if method == PiecewiseMethods.loglineardiscount:
        t_structure = ql.PiecewiseLogLinearDiscount(vdate,
                                                rate_helpers,
                                                ql_day_count[par_setting.day_count])
    elif method == PiecewiseMethods.linearforward:
        t_structure = ql.PiecewiseLinearForward(vdate,
                                                rate_helpers,
                                                ql_day_count[par_setting.day_count])
    elif method == PiecewiseMethods.linearzero:
        t_structure = ql.PiecewiseLinearZero(vdate,
                                            rate_helpers,
                                            ql_day_count[par_setting.day_count])
    elif method == PiecewiseMethods.cubiczero:
        t_structure = ql.PiecewiseCubicZero(vdate,
                                            rate_helpers,
                                            ql_day_count[par_setting.day_count])
    elif method == PiecewiseMethods.splinecubicdiscount:
        t_structure = ql.PiecewiseSplineCubicDiscount(vdate,
                                                rate_helpers,
                                                ql_day_count[par_setting.day_count])
    else:
        t_structure = ql.PiecewiseLogCubicDiscount(vdate,
                                                rate_helpers,
                                                ql_day_count[par_setting.day_count])
    #return the term structure
    return t_structure


def calc_curve(value_date: str, 
                depo_setting: DepoSetting, 
                depo_rates: List[Rate], 
                par_setting:BondSetting, 
                par_rates: List[Rate], 
                country = None, 
                method: str = "LogCubicDiscount" ):
        
    if country:
        try:
            calendar = ql_calendar_market[country]
        except:
            calendar = None
    t_structure = curve_piecewise(value_date, 
                depo_setting, 
                depo_rates, 
                par_setting, 
                par_rates, 
                cal = calendar,
                method= method)

    day_count = ql_day_count[par_setting.day_count]
    tenors = ["1D", "1W", "2W", "3W"]
    tenors_monthly = [ "".join([str(no + 1),"M"]) for no in range(360)]
    all_tenors = tenors + tenors_monthly

    vdate = datestr_to_qldate(value_date)
    dates = []
    dfs = []
    rates = []
    days = []
    maxDate = t_structure.maxDate()
    for atenor in all_tenors:
        adate = ql.NullCalendar().advance(vdate, 
                                        ql.Period(atenor),     
                                        ql_business_day[par_setting.business_day], 
                                        False)
        if adate <= maxDate:
            day = adate - vdate
            yrs = day_count.yearFraction(vdate, adate)
            compounding = ql.Compounded
            freq = ql_frequency[par_setting.frequency]
            
            zero_rate = t_structure.zeroRate(yrs, compounding, freq)
            tenors.append(yrs)
            eq_rate = zero_rate.equivalentRate(day_count,
                                            compounding,
                                            freq,
                                            vdate,
                                            adate)
            therate = eq_rate.rate()
            df = t_structure.discount(yrs,True)
            rates.append(therate*100)
            dfs.append(df)
            dates.append(adate.ISO())
            days.append(day)
        else:
            break

    zero_set = {"value_date": value_date, "day_count": par_setting.day_count, 
                "compound": "Compounded", "frequency": par_setting.frequency,
                "days": days, "dates": dates, "rates": rates}
    zerocurve = ZeroCurve(**zero_set)

    df_set = {"value_date": value_date, "day_count": par_setting.day_count, 
            "days": days, "dates": dates, "dfs": dfs}
    discountcurve = DiscountCurve(**df_set)

    return zero_set, df_set
    

def calc_qlzerorate(from_date: str, to_date: str, day_count: DayCount,
                    t_structure: ql.QuantLib.TermStructure, compounding,
                    frequency: Frequency):

    fdate = datestr_to_qldate(from_date)
    tdate = datestr_to_qldate(to_date)
    ql_dc = ql_day_count[day_count]
    yrs = ql_dc.yearFraction(fdate, tdate)
    zero_rate = t_structure.zeroRate(yrs, 
                                    ql_compounding[compounding], 
                                    ql_frequency[frequency])
    return zero_rate


def calc_equiv_rate(from_date: str, 
                    to_date: str,
                    qlrate: ql.QuantLib.InterestRate,
                    day_count: DayCount,
                    compounding,
                    frequency: Frequency):
    fdate = datestr_to_qldate(from_date)
    tdate = datestr_to_qldate(to_date)
    rate = qlrate.equivalentRate(ql_day_count[day_count],
                                ql_compounding[compounding],
                                ql_frequency[frequency],
                                fdate,
                                tdate).rate()
    return rate

    
def ql_spread_handle(value_date: str, 
                    t_structure: ql.QuantLib.YieldTermStructure, 
                    spreads: Optional[List[Rate]] = None,
                    calendar = None):
    vdate = datestr_to_qldate(value_date)
    ts_handle = ql.YieldTermStructureHandle(t_structure)
    
    if calendar == None:
        qlcalendar = ql.WeekendsOnly()
    else:
        qlcalendar = calendar

    spread_handles = []
    spread_dates = []
    spread_quotes = []
    handle = None
    if spreads:
        for spread in spreads:
            spread_quote = ql.SimpleQuote(spread.rate/100)
            spread_quotes.append(spread_quote)
            spread_handle = ql.QuoteHandle(spread_quote)
            spread_handles.append(spread_handle)
            maturity = qlcalendar.advance(vdate, get_qlPeriod(spread.tenor))
            spread_dates.append(maturity)

        spread_curve = ql.SpreadedLinearZeroInterpolatedTermStructure(
                                ts_handle,
                                spread_handles,
                                spread_dates)
        handle = ql.YieldTermStructureHandle(spread_curve)
    else:
        handle = ql.YieldTermStructureHandle
    

    return handle
    

def qlFlatForward(flatforward: FlatForwardCurve):
    qlDate = datestr_to_qldate(flatforward.date)
    qlHandle = ql.QuoteHandle(ql.SimpleQuote(flatforward.rate/100))
    qlDayCounter = ql_day_count[flatforward.day_count]
    qlCompounding = ql_compounding[flatforward.compounding]
    qlFrequency = ql_frequency[flatforward.frequency]
    
    return ql.FlatForward(qlDate, qlHandle, qlDayCounter, qlCompounding, qlFrequency)


def qlPiecewise(pwc: PiecewiseCurve, cal = None):
    return curve_piecewise(pwc.value_date, 
                            pwc.depo_setting, 
                            pwc.depo_rates, 
                            pwc.par_setting, 
                            pwc.par_rates,
                            cal = cal)


def qlYtsHandle(value_date: str, 
                t_structure: ql.QuantLib.YieldTermStructure, 
                spreads: Optional[List[Rate]] = None,
                calendar = None):
    return ql_spread_handle(value_date, 
                    t_structure, 
                    spreads,
                    calendar = calendar)





from QuantLib.QuantLib import OptionletVolatilityStructureHandle
from fpql.termstructure import qlFlatForward, qlPiecewise
import QuantLib as ql 
from .pymodels import *
from .ql_utils import *
from .ql_conventions import *
from .ql_enums import *
from . import pricingmodels as pm


def get_qlfixedratebond(bond_info: FixedRateBond, setting: BondSetting,
                calendar):
    
    ql_period = ql.Period(ql_freq_tenor[setting.frequency])
    schedule = ql.Schedule(datestr_to_qldate(bond_info.issue_date), 
                            datestr_to_qldate(bond_info.maturity), 
                            ql_period, 
                            calendar, 
                            ql_business_day[setting.business_day],
                            ql_business_day[setting.terminate_business_day],
                            ql_date_generation[setting.date_gen],
                            setting.is_eom)
    qlbond = ql.FixedRateBond(setting.start_basis, 
                                bond_info.face_value, 
                                schedule, [bond_info.coupon/100], 
                                ql_day_count[setting.day_count])
    return qlbond


def get_qlbond(bond_info: FixedRateBond, setting: BondSetting,
                calendar):
    ql_period = ql.Period(ql_freq_tenor[setting.frequency])
    day_count = ql_day_count[setting.day_count]
    schedule = ql.Schedule(datestr_to_qldate(bond_info.issue_date), 
                            datestr_to_qldate(bond_info.maturity), 
                            ql_period, 
                            calendar, 
                            ql_business_day[setting.business_day],
                            ql_business_day[setting.terminate_business_day],
                            ql_date_generation[setting.date_gen],
                            setting.is_eom)

    interest = ql.FixedRateLeg(schedule, 
                                day_count, 
                                [100.], 
                                [bond_info.coupon/100])
    qlbond = ql.Bond(0, ql.TARGET(), 
                    datestr_to_qldate(bond_info.issue_date), 
                    interest)
    
    return qlbond
    

def calc_fixbond_structures(value_date: str, 
                            bond_info: FixedRateBond, 
                            setting: BondSetting,
                            calendar):
    #create Schedule
    ql_period = ql.Period(ql_freq_tenor[setting.frequency])
    schedule = ql.Schedule(datestr_to_qldate(bond_info.issue_date), 
                            datestr_to_qldate(bond_info.maturity), 
                            ql_period, 
                            calendar, 
                            ql_business_day[setting.business_day],
                            ql_business_day[setting.terminate_business_day],
                            ql_date_generation[setting.date_gen],
                            setting.is_eom)
    
    # a sequence of QuantLib.QuantLib.SimpleCashFlow
    interests = ql.FixedRateLeg(schedule, 
                            ql_day_count[setting.day_count],
                            [bond_info.face_value], 
                            [bond_info.coupon/100],
                            ql.Following)
    no_of_interests = len(interests)
    schedule = list(schedule)
    no_of_dates = len(schedule)
    startdates = schedule[:no_of_dates - 1]
    enddates = schedule[1: no_of_dates]
    structures = []
    for i in range(no_of_interests):
        fv_flow = 0
        if i == no_of_interests - 1:
            cashflow = bond_info.face_value + interests[i].amount()
            fv_flow = bond_info.face_value
        else :
            cashflow = interests[i].amount()
    
        thestructure = {"start_date": startdates[i].ISO(),
                        "end_date": enddates[i].ISO(),
                        "payment_date": interests[i].date().ISO(),
                        "face_value": bond_info.face_value,
                        "coupon": bond_info.coupon,
                        "face_value_flow": fv_flow,
                        "interest": interests[i].amount(),
                        "cashflow": cashflow} 
        
        structure = Structure(**thestructure)  
        structures.append(structure)  
    
    return structures


def fixbond_ytm(clean_price: float, 
                daycount: DayCount,
                frequency: Frequency,
                thebond: ql.QuantLib.Bond):
    
    day_count = ql_day_count[daycount]
    freq = ql_frequency[frequency]
    ytm = thebond.bondYield(clean_price, 
                                day_count,
                                ql.Compounded,
                                freq)
    return ytm * 100


def bond_functions_cashflows(qlvaluedate:ql.QuantLib.Date, 
                            qlbond: ql.QuantLib.FixedRateBond):
    previous_cashflow_date = ql.BondFunctions.previousCashFlowDate(qlbond, qlvaluedate)
    previous_cashflow_amount = ql.BondFunctions.previousCashFlowAmount(qlbond, qlvaluedate)
    next_cashflow_date = ql.BondFunctions.nextCashFlowDate(qlbond, qlvaluedate)
    next_cashflow_amount = ql.BondFunctions.nextCashFlowAmount(qlbond, qlvaluedate)
    cf_info = {"previous_cashflow_date": previous_cashflow_date,
                "previous_cashflow_amount": previous_cashflow_amount,
                "next_cashflow_date": next_cashflow_date,
                "next_cashflow_amount": next_cashflow_amount}
    cfi = CashFlowInspector(**cf_info)
    return cfi


def bond_functions_coupons(qlbond: ql.QuantLib.FixedRateBond):
    previous_coupon_rate = ql.BondFunctions.previousCouponRate(qlbond)
    next_coupon_rate = ql.BondFunctions.nextCouponRate(qlbond)
    accrual_start_date = ql.BondFunctions.accrualStartDate(qlbond).ISO()
    accrual_end_date = ql.BondFunctions.accrualEndDate(qlbond)
    accrual_period = ql.BondFunctions.accrualPeriod(qlbond)
    accrual_days = ql.BondFunctions.accrualDays(qlbond)
    accrued_period = ql.BondFunctions.accruedPeriod(qlbond)
    accrued_days = ql.BondFunctions.accruedDays(qlbond)
    accrued_amount = ql.BondFunctions.accruedAmount(qlbond)

    c_info = {"previous_coupon_rate": previous_coupon_rate,
                "next_coupon_rate": next_coupon_rate, 
                "accrual_start_date": accrual_start_date,
                "accrual_end_date": accrual_end_date,
                "accrual_period": accrual_period,
                "accrual_days": accrual_days,
                "accrued_period": accrued_period,
                "accrued_days": accrued_days,
                "accrued_amount": accrued_amount}
    ci = CashFlowInspector(**c_info)

    return ci


def bond_functions_risks(qlbond: ql.QuantLib.FixedRateBond,
                        i_rate: ql.QuantLib.InterestRate):
    accrued = ql.BondFunctions.accruedAmount(qlbond)
    accrued_period = ql.BondFunctions.accruedPeriod(qlbond)
    clean_price = ql.BondFunctions.cleanPrice(qlbond, i_rate)
    bps = ql.BondFunctions.bps(qlbond, i_rate)
    duration = ql.BondFunctions.duration(qlbond, i_rate, ql.Duration.Macaulay)
    modified_duration = ql.BondFunctions.duration(qlbond, i_rate, ql.Duration.Modified)
    convexity = ql.BondFunctions.convexity(qlbond, i_rate)
    basis_point_value = ql.BondFunctions.basisPointValue(qlbond, i_rate)
    yield_value_basis_point = ql.BondFunctions.yieldValueBasisPoint(qlbond, i_rate)
    stats = {"accrued": accrued,
            "accrued_period": accrued_period,
            "clean_price": clean_price,
            "bps": bps,
            "duration": duration,
            "modified_duration": modified_duration,
            "convexity": convexity,
            "basis_point_value": basis_point_value,
            "yield_value_basis_point": yield_value_basis_point}
    risks = BondRisk(**stats)
    return risks


def z_spread(clean_price: float,
            fixbond:ql.QuantLib.FixedRateBond, 
            yieldcurve: ql.QuantLib.TermStructure,
            setting: BondSetting):
    return ql.BondFunctions.zSpread(fixbond, 
            clean_price, yieldcurve, 
            ql_day_count[setting.day_count], 
            ql.Compounded, 
            ql_frequency[setting.frequency])


def get_call_schedule(call_prices: List[CallabilityPrice]):
    callability_schedule = ql.CallabilitySchedule()

    for call_price in call_prices:
        c_price  = ql.CallabilityPrice(call_price.price, call_price.price_type)
        calldate = datestr_to_qldate(call_price.date)
        callability_schedule.append(ql.Callability(c_price, 
                                                    call_price.option_type,
                                                    calldate)
                                    )
    return callability_schedule


def get_qlcallablebond(bond_info: FixedRateBond,
                        bond_schedule: ql.QuantLib.Schedule,
                        coupons: List[float],
                        call_schedule: ql.QuantLib.CallabilitySchedule,
                        settlement_days: int = 0):
    
    callbond = ql.CallableFixedRateBond(
                settlement_days, 
                bond_info.face_value,
                bond_schedule, 
                coupons, 
                ql_day_count[bond_info.settings.day_count],
                ql_business_day[bond_info.settings.business_day], 
                bond_info.face_value, 
                datestr_to_qldate(bond_info.issue_date),
                call_schedule)
    return callbond


def callbond_stats(**params):
                    #clean_price: float,
                    #engine: ql.QuantLib.TreeCallableFixedRateBondEngine,
                    #curve: Optional[ql.QuantLib.YieldTermStructureHandle] = None):
    
    vdate = datestr_to_qldate(params.get('value_date'))
    call_class = params.get('callable_bond_param')
    ql_calendar = params.get('qlcalendar')
    ql_callbond = qlCallableBond_from_Class(call_class, ql_calendar)

    if params.get("curve"):
        curve_handle = params['curve']
    else:
        if call_class.use_flat_curve:
            if call_class.flat_curve:
                ffwd = qlFlatForward(call_class.flat_curve)
                curve_handle = ql.YieldTermStructureHandle(ffwd)
        else:
            pwc = qlPiecewise(call_class.piecewise_curve, ql_calendar)
            curve_handle = ql.YieldTermStructureHandle(pwc)
    
    engine = params.get('engine')
    if engine:
        ql_callbond.setPricingEngine(engine)
    else:
        return {}
    
    call_risks = {}
    
    if params.get('clean_price'):
        c_price = params['clean_price']
    else:
        c_price = ql_callbond.cleanPrice()
        call_risks['Clean Price'] = c_price
    oas = None

    try:
        oas = ql_callbond.OAS(c_price, 
                            curve_handle,
                            ql_day_count[call_class.info.settings.day_count],
                            ql.Compounded,
                            ql_frequency[call_class.info.settings.frequency],
                            vdate) * 100
        call_risks["OAS"] =  oas
        
    except:
        pass

    
    bond_yield  = ql_callbond.bondYield(c_price,
                        ql_day_count[call_class.info.settings.day_count],
                        ql.Compounded,
                        ql_frequency[call_class.info.settings.frequency],
                        vdate) * 100
    call_risks["Yield"] = bond_yield
    
    #OAS/Spread is not supported for trees other than OneFactorModel
    if oas is not None:
        e_convex = ql_callbond.effectiveConvexity(oas/100, 
                                    curve_handle,
                                    ql_day_count[call_class.info.settings.day_count],
                                    ql.Compounded,
                                    ql_frequency[call_class.info.settings.frequency])
        call_risks["Effective Convexity"] = e_convex

        e_duration = ql_callbond.effectiveDuration(oas/100, 
                                curve_handle,
                                ql_day_count[call_class.info.settings.day_count],
                                ql.Compounded,
                                ql_frequency[call_class.info.settings.frequency])
        call_risks["Effective Duration"] = e_duration


    return call_risks


def qlCallableBond_from_Class(call_class: CallableBondParam,
                            qlcalendar):
    qlschedule = qlSchedule_from_Model(call_class.info, qlcalendar) 
    structure = structure_from_qlSchedule(call_class.info,
                                            qlschedule)
    call_class.structure = structure
    qlcall_schedule = qlCallabilitySchedule_from_class(call_class.call_structure)
    qlcallbond = get_qlcallablebond(call_class.info, 
                                    qlschedule, 
                                    [call_class.info.coupon/100], 
                                    qlcall_schedule )
    
    return qlcallbond

        


import QuantLib as ql
from .ql_enums import *
from .ql_utils import *
from .pymodels import *
from .ql_conventions import *


def create_fixratestructure(thebond: FixedRateBond,
                            ql_calendar ):
    setting = thebond.settings
    
    period = ql.Period(ql_freq_tenor[setting.frequency])

    schedule = qlSchedule(thebond.issue_date, 
                            thebond.maturity, 
                            period,
                            ql_calendar,
                            setting.business_day, 
                            setting.business_day,
                            setting.date_gen, 
                            eom= True)
    interests = ql.FixedRateLeg(schedule, 
                            ql_day_count[setting.day_count],
                            [thebond.face_value], 
                            [thebond.coupon/100])
    no_of_interests = len(interests)
    structures = []
    for i in range(no_of_interests):
        if i == no_of_interests - 1:
            thestructure = {"start_date": schedule[i].ISO(),
                            "end_date": interests[i].date().ISO(),
                            "face_value": thebond.face_value,
                            "coupon": thebond.coupon,
                            "face_value_flow": thebond.face_value,
                            "interest": interests[i].amount(),
                            "cashflow": thebond.face_value + interests[i].amount()} 
        else:
            thestructure = {"start_date": schedule[i].ISO(),
                            "end_date": interests[i].date().ISO(),
                            "face_value": thebond.face_value,
                            "coupon": thebond.coupon,
                            "interest": interests[i].amount(),
                            "cashflow": interests[i].amount()}
        structure = Structure(**thestructure)  
        structures.append(structure)  
    
    return structures


def testapis():
    print("TEST APIS")

    
    


import datetime

from .ql_enums import *
from .ql_conventions import *

from pydantic import BaseModel, validator
from typing import List, Optional
from numpy import datetime64 as dt64



class Rate(BaseModel):
    tenor: str
    rate: float

    @validator('tenor')
    def valid_tenor(cls, v):
        prd = v[-1]
        number = v[:len(v)-1]
        if prd not in ['d', 'D', 'w', 'W', 'm', 'M', "y", "Y"]:
            raise ValueError('not a valid period')
        else:
            try:
                x = int(number)
            except:
                raise ValueError('not a valid period')
        return  v

    @validator('rate')
    def valid_rate(cls, v):
        
        try:
            x = float(v)
        except:
            raise ValueError('rate must be a number')
        return  v


class Holiday(BaseModel):
    date: str

    @validator('date')
    def valid_rate(cls, v): 
        try:
            thedate = dt64(v)
        except:
            raise ValueError('not a valid date format for Holidays')
        return  v


class DepoSetting(BaseModel):
    start_basis: StartBasis = StartBasis.today
    business_day: BusinessDay = BusinessDay.no_adjustment
    day_count: DayCount = DayCount.act_act_act365
    eom: bool = False


class BondSetting(BaseModel):
    frequency: Frequency = Frequency.semi_annual
    start_basis: StartBasis = StartBasis.today
    day_count: DayCount = DayCount.act_act_act365
    #pmt_day_count: DayCount = DayCount.act_act_act365
    business_day: BusinessDay = BusinessDay.no_adjustment
    terminate_business_day: BusinessDay = BusinessDay.mod_following
    date_gen: DateGeneration = DateGeneration.backward
    is_eom: bool = False


    @validator('day_count')
    def valid_day_count(cls, v): 
        if v not in ql_day_count.keys():
            raise ValueError
        return  v

    @validator('business_day')
    def valid_business_day(cls, v): 
        if v not in ql_business_day.keys():
            raise ValueError
        return  v

    @validator('frequency')
    def valid_frequency(cls, v): 
        if v not in ql_frequency.keys():
            raise ValueError
        return  v

    @validator('date_gen')
    def valid_date_gen(cls, v): 
        if v not in ql_date_generation.keys():
            raise ValueError
        return  v


class Structure(BaseModel):
    start_date: str 
    end_date: str 
    payment_date: Optional[str] = None
    face_value: float 
    coupon: float 
    dcf: Optional[float]= None
    interest: Optional[float] = None
    face_value_flow: float = 0
    cashflow: Optional[float]= None

    @validator('start_date', 'end_date')
    def valid_issue_date(cls, v): 
        try:
            adate = dt64(v)
        except:
            raise ValueError
        return  v


    @validator('face_value', 'coupon', 'face_value_flow')
    def valid_float(cls, v): 
        try:
            anumber = float(v)
        except:
            raise ValueError
        return  v


class DiscountCurve(BaseModel):
    value_date: str
    day_count: DayCount
    dates: List[str] = []
    days: List[int] = []
    dfs: List[float] = []


    @validator('value_date')
    def valid_value_date(cls, v): 
        try:
            adate = dt64(v)
        except:
            raise ValueError
        return  v


class ZeroCurve(BaseModel):
    value_date: str
    day_count: DayCount
    compound: Compounding
    frequency: Frequency
    dates: List[str] = []
    days: List[int] = []
    rates: List[float] = []


    @validator('value_date')
    def valid_value_date(cls, v): 
        try:
            adate = dt64(v)
        except:
            raise ValueError
        return  v


    @validator('day_count')
    def valid_day_count(cls, v): 
        if v not in ql_day_count.keys():
            raise ValueError
        return  v


    @validator('compound')
    def valid_compound(cls, v): 
        if v not in ql_compounding.keys():
            raise ValueError
        return  v


    @validator('frequency')
    def valid_frequency(cls, v): 
        if v not in ql_frequency.keys():
            raise ValueError
        return  v


class FixedRateBond(BaseModel):
    issue_date: Optional[str] = None
    maturity: str
    settings: BondSetting
    face_value: Optional[float] = 100.00
    coupon: float 
    structure: Optional[List[Structure]] = None

    @validator('face_value', 'coupon')
    def valid_float(cls, v): 
        try:
            anumber = float(v)
        except:
            raise ValueError
        return  v

    @validator('issue_date', 'maturity')
    def valid_date(cls, v): 
        try:
            adate = dt64(v)
        except:
            raise ValueError
        return  v


class StructuredBond(BaseModel):
    issue_date: Optional[str] = None
    maturity: str
    settings: BondSetting
    structure: List[Structure] = []

    @validator('maturity')
    def valid_maturity(cls, v): 
        try:
            adate = dt64(v)
        except:
            raise ValueError
        return  v

    @validator('issue_date')
    def valid_issue_date(cls, v): 
        try:
            adate = dt64(v)
        except:
            raise ValueError
        return  v


class CouponInspector(BaseModel):
    previous_coupon_rate: float
    next_coupon_rate: float 
    accrual_start_date: str
    accrual_end_date: str
    accrual_period: float
    accrual_days: float
    accrued_period: float
    accrued_days: float
    accrued_amount: float


class CashFlowInspector(BaseModel):
    previous_cashflow_date: str
    previous_cashflow_amount: float
    next_cashflow_date: float
    next_cashflow_amount: float


class BondRisk(BaseModel):
    accrued: Optional[float] = None
    ytm: Optional[float] = None
    clean_price: Optional[float] = None
    bps: Optional[float] = None
    duration: Optional[float] = None
    modified_duration: Optional[float] = None
    convexity: Optional[float] = None
    basis_point_value: Optional[float] = None
    yield_value_basis_point: Optional[float] = None
    

class BondRisk2(BondRisk):
    value: Optional[float] = None
    value_after_shift: Optional[float] = None


class CashFlow(BaseModel):
    date: str 
    amount: float

    @validator('date')
    def valid_date(cls, v): 
        try:
            adate = dt64(v)
        except:
            raise ValueError
        return  v

    @validator('amount')
    def valid_float(cls, v): 
        try:
            anumber = float(v)
        except:
            raise ValueError
        return  v


class CallabilityPrice(BaseModel):
    price_type: BondPriceType = BondPriceType.clean
    price: float
    date: str
    option_type: OptionType = OptionType.calloption


class FlatForwardCurve(BaseModel):
    date: str
    rate: float
    day_count: DayCount
    frequency: Frequency
    compounding: Optional[Compounding] = "Compounded" 


class PiecewiseCurve(BaseModel):
    value_date: str
    depo_setting: DepoSetting
    depo_rates: List[Rate]
    par_setting: BondSetting
    par_rates: List[Rate]
    method: PiecewiseMethods = PiecewiseMethods.logcubicdiscount


class CallableRisks(BaseModel):
    oas: Optional[float] = None
    bond_yield: Optional[float] = None
    ytc: Optional[float] = None
    eff_convexity: Optional[float] = None
    eff_duration: Optional[float] = None


class CallableBondParam(BaseModel):
    flat_curve: Optional[FlatForwardCurve] = None
    piecewise_curve:Optional[PiecewiseCurve] = None
    info: FixedRateBond
    structure: List[Structure] = []
    call_structure: List[CallabilityPrice] = []
    risks: Optional[CallableRisks] = None
    irmodel: Optional[ShortRateModel] = ShortRateModel.hullwhite
    use_flat_curve: bool = True


class SRMVasicek(BaseModel):
    r0: float = 0.05
    a: float = 0.1
    b: float = 0.05
    sigma: float = 0.1
    lamda: float = 0.0


class SRMBlackKaransinski(BaseModel):
    a: float = 0.1
    sigma: float = 0.1


class SRMHullWhite(BaseModel):
    a: float = 0.1
    sigma: float = 0.01


class SRMG2(BaseModel):
    a: float = 0.1
    b: float = 0.1
    sigma: float = 0.01
    eta: float = 0.01
    rho: float = -0.75



import QuantLib as ql
from typing import Optional
from . import ql_enums as qle

def get_qlModel(model: qle.ShortRateModel,
                ts_handle: Optional[ql.QuantLib.YieldTermStructureHandle]=None,
                r0=0.05, 
                a=0.1, 
                b=0.05, 
                sigma=0.01, 
                lamda=0.0,
                eta=0.01, 
                rho=-0.75,
                volstepdates = None, 
                volatilities = None, 
                reversions = None):

    if model == qle.ShortRateModel.vasicek: 
        return ql.Vasicek(r0, a, b, sigma, lamda)
    elif model == qle.ShortRateModel.blackkarasinski:
        return ql.BlackKarasinski(ts_handle, a, sigma)
    elif model == qle.ShortRateModel.hullwhite:
        return ql.HullWhite(ts_handle, a, sigma)
    elif model == qle.ShortRateModel.gsr:
        return ql.Gsr(ts_handle, volstepdates, volatilities, reversions)
    elif model == qle.ShortRateModel.g2:
        return ql.G2(ts_handle, a, sigma, b, eta, rho) 
    else:
        return None


def qlShortRateModel(srmodel: qle.ShortRateModel,
                    yts_handle: ql.QuantLib.YieldTermStructureHandle,
                    **params):

    if srmodel == qle.ShortRateModel.vasicek: 
        return ql.Vasicek(params["r0"], params["a"], params["b"], params["sigma"], params["lamda"])
    elif srmodel == qle.ShortRateModel.blackkarasinski:
        return ql.BlackKarasinski(yts_handle, params["a"], params["sigma"])
    elif srmodel == qle.ShortRateModel.hullwhite:
        return ql.HullWhite(yts_handle, params["a"], params["sigma"])
    elif srmodel == qle.ShortRateModel.gsr:
        return ql.Gsr(yts_handle, params["volstepdates"], params["volatilities"], params["reversions"])
    elif srmodel == qle.ShortRateModel.g2:
        return ql.G2(yts_handle, params["a"], params["sigma"], params["b"], params["eta"], params["rho"]) 
    else:
        return None
    

def qlTreeCallableEngine(srmodel, gridpoints: int = 100):
    return ql.TreeCallableFixedRateBondEngine(srmodel, gridpoints)







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import QuantLib as ql

"""
Created on Tue Jul  18 17:33:15 2021

@author: RMS671214
"""

ql_day_count = {
    'Actual/365 Fixed': ql.Actual365Fixed(), 
    'Actual/365 Fixed (Canadian)': ql.Actual365Fixed(ql.Actual365Fixed.Canadian), 
    'Actual/365 Fixed (No Leap)': ql.Actual365Fixed(ql.Actual365Fixed.NoLeap),
    'Actual/360': ql.Actual360(),
    'Actual/Actual': ql.ActualActual(),
    'Actual/Actual (ISMA)':ql.ActualActual(ql.ActualActual.ISMA),
    'Actual/Actual (Bond)':ql.ActualActual(ql.ActualActual.Bond),
    'Actual/Actual (ISDA)':ql.ActualActual(ql.ActualActual.ISDA),
    'Actual/Actual (Historical)':ql.ActualActual(ql.ActualActual.Historical),
    'Actual/Actual (Actual365)':ql.ActualActual(ql.ActualActual.Actual365),
    'Actual/Actual (AFB)':ql.ActualActual(ql.ActualActual.AFB),
    'Business252':ql.Business252(),
    'Thirty360': ql.Thirty360()
}  

ql_business_day = {
    "No Adjustment": ql.Unadjusted, 
    "Following": ql.Following, 
    "Modified Following": ql.ModifiedFollowing,
    "Preceding": ql.Preceding, 
    "Modified Preceding": ql.ModifiedPreceding
}

ql_date_generation = {
    "Forward from issue date": ql.DateGeneration.Forward,
    "Backward from maturity date": ql.DateGeneration.Backward,
    "Zero": ql.DateGeneration.Zero,
    "ThirdWednesday": ql.DateGeneration.ThirdWednesday,
    "Twentieth": ql.DateGeneration.Twentieth,
    "TwentiethIMM": ql.DateGeneration.TwentiethIMM,
    "CDS": ql.DateGeneration.CDS,
}

ql_frequency = {
    "No Frequency": ql.NoFrequency,
    "Once": ql.Once,
    "Annual": ql.Annual, 
    "Semi-Annual": ql.Semiannual, 
    "Every Four Months": ql.EveryFourthMonth,
    "Quarterly": ql.Quarterly,
    "Bi-Monthly": ql.Bimonthly,
    "Monthly": ql.Monthly,
    "Every Fourth Week": ql.EveryFourthWeek,
    "Bi-Weekly": ql.Biweekly,
    "Weekly": ql.Weekly,
    "Daily": ql.Daily
}

ql_freq_tenor = {
    "Annual": "1Y",
    "Semi-Annual": "6M",
    "Every Four Months": "4M",
    "Quarterly": "3M",
    "Bi-Monthly": "2M",
    "Monthly": "1M",
    "Every Fourth Week": "4W",
    "Bi-Weekly": "2W",
    "Weekly": "1W",
    "Daily": "1D"
}

ql_tenor = {
    "M": ql.Months,
    "Y": ql.Years,
    "W": ql.Weeks,
    "D": ql.Days
}

ql_compounding = {
    "Simple": ql.Simple,
    "Compounded": ql.Compounded,
    "Continuous": ql.Continuous,
    "SimpleThenCompounded": ql.SimpleThenCompounded,
    "CompoundedThenSimple": ql.CompoundedThenSimple
}

ql_calendar_market = {
    "Argentina": ql.Argentina(),
    "Argentina (Merval)": ql.Argentina(ql.Argentina.Merval),
    "Brazil" : ql.Brazil(),
    "Brazil (Exchange)" : ql.Brazil(ql.Brazil.Exchange),
    "Brazil (Settlement)" : ql.Brazil(ql.Brazil.Settlement),
    "Canada": ql.Canada(),
    "Canada (Settlement)": ql.Canada(ql.Canada.Settlement),
    "Canada (TSX)": ql.Canada(ql.Canada.TSX),
    "China": ql.China(),
    "China (IB)": ql.China(ql.China.IB),
    "China (SSE)": ql.China(ql.China.SSE),
    "Czech Republic" : ql.CzechRepublic(),
    "Czech Republic (PSE)" : ql.CzechRepublic(ql.CzechRepublic.PSE),
    "France" : ql.France(),
    "France (Exchange)" : ql.France(ql.France.Exchange),
    "France (Settlement)" : ql.France(ql.France.Settlement),
    "Germany" : ql.Germany(),
    "Germany (Eurex)": ql.Germany(ql.Germany.Eurex),
    "Germany (FrankfurtStockExchange)": ql.Germany(ql.Germany.FrankfurtStockExchange), 
    "Germany (Settlement)": ql.Germany(ql.Germany.Settlement),
    "Germany (Xetra)": ql.Germany(ql.Germany.Xetra),
    "Hong Kong": ql.HongKong(),
    "Hong Kong (HKEx)": ql.HongKong(ql.HongKong.HKEx),
    "Iceland": ql.Iceland(),
    "Iceland (ICEX)" : ql.Iceland(ql.Iceland.ICEX),
    "India" : ql.India(),
    "India (NSE)": ql.India(ql.India.NSE),
    "Indonesia" : ql.Indonesia(),
    "Indonesia (BEJ)": ql.Indonesia(ql.Indonesia.BEJ),
    "Indonesia (JSX)": ql.Indonesia(ql.Indonesia.JSX),
    "Israel": ql.Israel(),
    "Israel (Settlement)": ql.Israel(ql.Israel.Settlement),
    "Israel (TASE)": ql.Israel(ql.Israel.TASE),
    "Italy": ql.Italy(),
    "Italy (Exchange)": ql.Italy(ql.Italy.Exchange),
    "Italy (Settlement)": ql.Italy(ql.Italy.Settlement),
    "Mexico": ql.Mexico(),
    "Mexico (BMV)": ql.Mexico(ql.Mexico.BMV),
    "Russia": ql.Russia(),
    "Russia (MOEX)": ql.Russia(ql.Russia.MOEX),
    "Russia (Settlement)": ql.Russia(ql.Russia.Settlement),
    "Saudi Arabia": ql.SaudiArabia(),
    "Saudi Arabia (Tadawul)": ql.SaudiArabia(ql.SaudiArabia.Tadawul),
    "Singapore": ql.Singapore(),
    "Singapore (SGX)": ql.Singapore(ql.Singapore.SGX),
    "Slovakia": ql.Slovakia(),
    "Slovakia (BSSE)": ql.Slovakia(ql.Slovakia.BSSE),
    "South Korea": ql.SouthKorea(),
    "South Korea (KRX)": ql.SouthKorea(ql.SouthKorea.KRX),
    "South Korea (Settlement)": ql.SouthKorea(ql.SouthKorea.Settlement),
    "Taiwan": ql.Taiwan(),
    "Taiwan (TSEC)": ql.Taiwan(ql.Taiwan.TSEC),
    "Ukraine" : ql.Ukraine(),
    "Ukraine (USE)": ql.Ukraine(ql.Ukraine.USE),
    "United Kingdom" : ql.UnitedKingdom(),
    "United Kingdom (Exchange)": ql.UnitedKingdom(ql.UnitedKingdom.Exchange),
    "United Kingdom (Metals)": ql.UnitedKingdom(ql.UnitedKingdom.Metals),
    "United Kingdom (Settlement)": ql.UnitedKingdom(ql.UnitedKingdom.Settlement),
    "United States": ql.UnitedStates(),
    "United States (FederalReserve)": ql.UnitedStates(ql.UnitedStates.FederalReserve),
    "United States (GovernmentBond)": ql.UnitedStates(ql.UnitedStates.GovernmentBond),
    "United States (LiborImpact)": ql.UnitedStates(ql.UnitedStates.LiborImpact),
    "United States (NREC)": ql.UnitedStates(ql.UnitedStates.NERC),
    "United States (NYSE)": ql.UnitedStates(ql.UnitedStates.NYSE),
    "United States (Settlement)": ql.UnitedStates(ql.UnitedStates.Settlement) ,
    "TARGET": ql.TARGET(),
    "WeekendsOnly": ql.WeekendsOnly(),
    "NullCalendar": ql.NullCalendar()
}

ql_start_basis = { "Today": 0, "Tom": 1, "Spot": 2}


irmodels = ['Vasicek', 'BlackKarasinski', 'HullWhite', "gsr", 'g2']





from typing import List, Optional
from .pymodels import *
from .ql_conventions import *
import QuantLib as ql


def create_calendar(description: Optional[str], 
                    holidays:List[Holiday] = [], 
                    mycal = None):

    cal = ql.BespokeCalendar(description)
    qldates = [datestr_to_qldate(holiday.date) for holiday in holidays]
    for qldate in qldates:
        cal.addHoliday(qldate)
    
    if not mycal:
        calendar = ql.JointCalendar(cal, ql.WeekendsOnly())
    else:
        calendar = ql.JointCalendar(cal, mycal)
    
    return calendar
    

def datestr_to_qldate(datestr):
    vdate = datestr.split("-")
    vdt = [int(vdat) for vdat in vdate]
    return ql.Date(vdt[2], vdt[1], vdt[0])


def qlSchedule_from_Model(info: FixedRateBond, 
                        ql_calendar):
    return qlSchedule(info.issue_date,
                        info.maturity,
                        ql.Period(ql_frequency[info.settings.frequency]), 
                        ql_calendar, 
                        info.settings.business_day, 
                        info.settings.business_day, 
                        info.settings.date_gen, 
                        info.settings.is_eom)


def structure_from_qlSchedule(bond_info: FixedRateBond,
                            schedule: ql.QuantLib.Schedule):
    interests = ql.FixedRateLeg(schedule, 
                            ql_day_count[bond_info.settings.day_count],
                            [bond_info.face_value], 
                            [bond_info.coupon/100],
                            ql.Following)
    no_of_interests = len(interests)
    schedule = list(schedule)
    no_of_dates = len(schedule)
    startdates = schedule[:no_of_dates - 1]
    enddates = schedule[1: no_of_dates]
    structures = []
    for i in range(no_of_interests):
        if i == no_of_interests - 1:
            cashflow = bond_info.face_value + interests[i].amount()
        else :
            cashflow = interests[i].amount()
        thestructure = {"start_date": startdates[i].ISO(),
                        "end_date": enddates[i].ISO(),
                        "payment_date": interests[i].date().ISO(),
                        "face_value": bond_info.face_value,
                        "coupon": bond_info.coupon,
                        "face_value_flow": bond_info.face_value,
                        "interest": interests[i].amount(),
                        "cashflow": cashflow} 
        
        structure = Structure(**thestructure)  
        structures.append(structure)  
    return structures


def qlSchedule_from_Structure(structures: List[Structure]):

    dates = []
    dates.append(structures[0].start_date)
    for structure in structures:
        dates.append(structure.end_date)

    qlschedule = ql.Schedule([datestr_to_qldate(dt) for dt in dates])
    return qlschedule


def get_qlPeriod(tenor:str):
    prd = tenor[-1]
    number = int(tenor[:len(tenor)-1])
    return  ql.Period(number, ql_tenor[prd])


def qlSchedule(issue_date: str, 
            maturity: str, 
            period: ql.QuantLib.Period, 
            ql_calendar,
            business_day: BusinessDay, 
            maturity_business_day: BusinessDay,
            date_gen: DateGeneration, 
            eom: bool = False):
    """
    Creates list of dates. 

        Parameters:
            issue_date (str): if date_gen is DateGeneration.backward, issue_date can 
                be treated as the value date, otherwise please use issue date for 
                accurate dates.
            maturity (str): maturity of the instrument.
            frequency (str): frequency of coupon payment
            ql_calendar: ql.Calendar relevant for the currency.
            business_day (str): business day convention for the dates.
            maturity_business_day (str): business day convention for the 
                maturity of the instrument
            date_gen (str): manner in which dates are generated.
            eom (bool): if the start date is at the end of the month, whether 
                other dates are required to be scheduled at the end of the month
                (except the last date).

        Returns:
            list of ql.Dates()
    """
    vdate = datestr_to_qldate(issue_date)
    mdate = datestr_to_qldate(maturity)
    
    schedule = ql.Schedule(vdate, 
                mdate, 
                period, 
                ql_calendar, 
                ql_business_day[business_day],
                ql_business_day[maturity_business_day], 
                ql_date_generation[date_gen], 
                eom)
    return schedule


def qlMakeScedule(issue_date: str, maturity: str, frequency:str, ql_calendar = None,
                business_day: Optional[BusinessDay] = None, 
                maturity_business_day: Optional[BusinessDay] = None,
                date_gen: Optional[DateGeneration] = None, eom: bool = False):

    vdate = datestr_to_qldate(issue_date)
    mdate = datestr_to_qldate(maturity)
    freq = ql.Period(frequency)
    if business_day:
        ql_bus_day = ql_business_day[business_day]
    if maturity_business_day:
        ql_mat_bus_day = ql_business_day[maturity_business_day]
    if date_gen:
        ql_date_gen = ql_date_generation[date_gen]
    schedule = ql.MakeSchedule(vdate, mdate, freq)
    return schedule


def qlCallabilitySchedule_from_class(call_schedule: List[CallabilityPrice]):
    callability_schedule = ql.CallabilitySchedule()
    
    for calldata in call_schedule:
        call_price  = ql.CallabilityPrice(calldata.price, 
                                            calldata.price_type)
        call_date = datestr_to_qldate(calldata.date)
        callability = ql.Callability(call_price, 
                                calldata.option_type,
                                call_date)
        callability_schedule.append(callability)
    return callability_schedule



from pydantic import BaseModel, validator
from typing import List
from enum import Enum


class StartBasis(int, Enum):
    spot = 2
    tom = 1
    today = 0


class BusinessDay(str, Enum):
    no_adjustment = "No Adjustment" 
    following = "Following" 
    mod_following = "Modified Following"
    preceding = "Preceding"
    mod_preceding = "Modified Preceding"


class DayCount(str, Enum):
    act_365_fixed = 'Actual/365 Fixed' 
    act_365_fixed_canadian = 'Actual/365 Fixed (Canadian)'
    act_365_fixed_noleap = 'Actual/365 Fixed (No Leap)'
    act_360 = 'Actual/360'
    act_act = 'Actual/Actual'
    act_act_isma = 'Actual/Actual (ISMA)'
    act_act_bond = 'Actual/Actual (Bond)'
    act_act_isda = 'Actual/Actual (ISDA)'
    act_act_historical = 'Actual/Actual (Historical)'
    act_act_act365 ='Actual/Actual (Actual365)'
    act_act_afb = 'Actual/Actual (AFB)'
    business252 = 'Business252'
    thirty360 = 'Thirty360'


class DateGeneration(str, Enum):
    forward = "Forward from issue date"
    backward = "Backward from maturity date"
    zero = "Zero"
    third_wednesday = "ThirdWednesday"
    twentieth = "Twentieth"
    twentieth_IMM = "TwentiethIMM"
    cds = "CDS"


class Frequency(str, Enum):
    no_frequency = "No Frequency"
    once = "Once"
    annual = "Annual" 
    semi_annual = "Semi-Annual"
    every_four_months = "Every Four Months"
    quarterly = "Quarterly"
    bi_monthly = "Bi-Monthly"
    monthly = "Monthly"
    every_fourth_week = "Every Fourth Week"
    bi_weekly = "Bi-Weekly"
    weekly = "Weekly"
    daily = "Daily"


class Country(str, Enum):
    argentina = "Argentina"
    brazil = "Brazil"
    canada = "Canada"
    china = "China"
    czechRepublic  = "Czech Republic" 
    france = "France" 
    germany = "Germany" 
    hongkong = "Hong Kong"
    iceland = "Iceland"
    india = "India" 
    indonesia = "Indonesia" 
    israel = "Israel"
    italy = "Italy"
    mexico = "Mexico"
    russia = "Russia"
    saudiarabia = "Saudi Arabia"
    singapore = "Singapore"
    slovakia = "Slovakia"
    southkorea = "South Korea"
    taiwan = "Taiwan"
    ukraine = "Ukraine" 
    unitedkingdom = "United Kingdom" 
    unitedstates = "United States"
    target = "TARGET"


class PiecewiseMethods(str, Enum):
    logcubicdiscount = "LogCubicDiscount"
    loglineardiscount = "LogLinearDiscount"
    linearzero = "LinearZero"
    cubiczero = "CubicZero"
    linearforward = "LinearForward"
    splinecubicdiscount = "SplineCubicDiscount"


class ZeroCurveMethods(str, Enum):
    zerocurve = "ZeroCurve"
    loglinearZeroCurve = "LogLinearZeroCurve"
    cubiczerocurve = "CubicZeroCurve"
    naturalcubiczerocurve = "NaturalCubicZeroCurve"
    logcubiczerocurve = "LogCubicZeroCurve"
    monotoniccubiczerocurve = "MonotonicCubicZeroCurve"


class Compounding(str, Enum):
    simple = "Simple"
    compounded = "Compounded"
    continuous = "Continuous"
    simplethencontinuous = "SimpleThenCompounded"
    compoundedthensimple = "CompoundedThenSimple"


class ShortRateModel(str, Enum):
    vasicek = 'Vasicek'
    blackkarasinski = 'BlackKarasinski'
    hullwhite = 'HullWhite'
    gsr = "gsr"
    g2 = 'g2'


class BondPriceType(int, Enum):
    dirty = 0
    clean = 1


class OptionType(int,Enum):
    calloption = 0
    putoption = 1

